import pandas as pd
import numpy as np
from sklearn.ensemble import RandomForestRegressor
from sklearn.model_selection import KFold
from sklearn.metrics import mean_squared_error
import matplotlib.pyplot as plt
import seaborn as sns
import joblib
import os
from sklearn.metrics import mean_squared_error, r2_score
from sklearn.preprocessing import OneHotEncoder

# ----------------------------------------------------------- #
#### Functions ####
# ----------------------------------------------------------- #
# Set the seed for reproducibility
np.random.seed(666)

def k_fold_CV(df_keys, K, Ptst=None):
      # df_keys: data frame with id and Y
    # K: number of folds
    # Ptst: fraction of the test set, of the number of genotypes (not rows in df_keys)
    I = df_keys['Exp_ID'].unique()#12 unique indivdueal
    J = len(I)# unique individuals (I), and number (J)
    Ntst = int(J // K) if Ptst is None else  (J * Ptst) #size of test set for 1 fold ifelse(is.null(Ptst), floor(J/K), J*Ptst)
     # initialise empty fold list and fill I_rest
    l_ID_all = []
    I_rest = I.copy()

    for k in range(K):
        if len(I_rest) >= Ntst:# if there's more to sample than to be sampled for the next fold
            I_fold = np.random.choice(I_rest, round(Ntst), replace=False)
            l_ID_all.append(I_fold) # append sampled fold to list
            I_rest = np.setdiff1d(I_rest, I_fold) # subtract from I_rest
        else:# not enough left to sample in I_rest
            I_fold = I_rest # add all of I_rest to fold
            I_rest = I.copy()   # start new I_rest
            I_fold_add = np.random.choice(I_rest, round(Ntst - len(I_fold)), replace=False) # complete the sampling
            I_rest = np.setdiff1d(I_rest, I_fold_add)# subtract from I_rest and complete I_fold
            I_fold = np.concatenate([I_fold, I_fold_add])
            l_ID_all.append(I_fold) # append sampled fold to list

    l_res = [df_keys[df_keys['Exp_ID'].isin(fold)] for fold in l_ID_all]
    return l_res

def predict_RF_lnp(df_arr_res, pred_param, var):
      # df_arr_res: df from SIM files
    # pred_param: LNP column to boost (NFGMR, THARV or TAGBG), and compare eventual RF-prediction to
    df_keys = df_arr_res[['Exp_ID', 'Year']].drop_duplicates()

    l_test = k_fold_CV(df_keys, K=10, Ptst=0.35)
    l_test_ind = {
    f'fold_{i}': df_arr_res.reset_index().merge(test_keys, on=['Exp_ID', 'Year'])['index'].tolist()
    for i, test_keys in enumerate(l_test)
    }
    df_features = df_arr_res.drop(columns=['Exp_ID', 'Year', 'DATE', 'OBSERVATIONS','PUSHPL','RCO','RRDD',"graslandleeftijd","textuur"]) 
    resp_var = df_arr_res['OBSERVATIONS']

    l_m_DMY = [RandomForestRegressor(n_estimators=1000, max_features=153).fit(df_features.drop(index=l_test_ind.get(key)), resp_var.drop(index=l_test_ind.get(key))) for key in l_test_ind.keys()]
    
    # Save the trained models
    # Create directory if it doesn't exist
    current_path = os.getcwd()
    model_path = os.path.join(current_path, "grassat/input/random_forest_training/trainedRFmodels"+var)
    if not os.path.exists(model_path):
        os.makedirs(model_path)

    for i, model in enumerate(l_m_DMY):
        joblib.dump(model, os.path.join(model_path, f'random_forest_model_{i}.pkl'))
    
    l_predictions = []

    for i, test_ind in enumerate(l_test_ind):
        preds = l_m_DMY[i].predict(df_features.loc[l_test_ind.get(test_ind)])
        obs = resp_var.loc[l_test_ind.get(test_ind)]
        l_predictions.append(pd.DataFrame({
            'Exp_ID': df_arr_res.loc[l_test_ind.get(test_ind), 'Exp_ID'],
            'Year': df_arr_res.loc[l_test_ind.get(test_ind), 'Year'],
            'DOY': df_arr_res.loc[l_test_ind.get(test_ind), 'DOY'],
            pred_param: df_arr_res.loc[l_test_ind.get(test_ind), pred_param],
            'prediction': preds,
            'observations': obs
        }))

    df_predict = pd.concat(l_predictions).groupby(['Year', 'Exp_ID', 'observations', pred_param]).agg(
        sd=('prediction', 'std'),
        prediction=('prediction', 'mean')
    ).reset_index()

    return df_predict

#evaluation metrics

def evaluation_metrics(df_predict, outdir):
    sns.lmplot(data=df_predict, x='observations', y='value', hue='Year', col='variable', fit_reg=True)
    plt.savefig(outdir+'/plot.png')
    plt.show()

    # Calculate evaluation statistics
    evaluation_stats_list = []

    for model in df_predict['variable'].unique():
        df_model = df_predict[df_predict['variable'] == model]
        mse = mean_squared_error(df_model['observations'], df_model['value'])
        rmse = mean_squared_error(df_model['observations'], df_model['value'], squared=False)
        r2 = r2_score(df_model['observations'], df_model['value'])
        
        evaluation_stats_list.append({
            'Model': model,
            'Metric': 'Mean Squared Error (MSE)',
            'Value': mse
        })
        evaluation_stats_list.append({
            'Model': model,
            'Metric': 'Root Mean Squared Error (RMSE)',
            'Value': rmse
        })
        evaluation_stats_list.append({
            'Model': model,
            'Metric': 'R-squared (R2)',
            'Value': r2
        })

    # Convert the dictionary to a DataFrame
    df_evaluation_stats = pd.DataFrame(evaluation_stats_list)

    # Save the DataFrame to a CSV file
    df_evaluation_stats.to_csv(outdir+'/statistics.csv', index=False)
    

# ----------------------------------------------------------- #
#### Prediction ####
# ----------------------------------------------------------- #

# Crude protein
current_path = os.getcwd()
CP_path = os.path.join(current_path, "grassat/input/random_forest_training/SIM_CP_NFGMR.csv")#
df_SIM_NFGMR = pd.read_csv(CP_path)

texture=False
# Fit and transform the categorical feature
if texture==True:
    encoder = OneHotEncoder()
    encoded_features = encoder.fit_transform(df_SIM_NFGMR[['textuur']]).toarray()
    # Convert the encoded features to a DataFrame
    encoded_df = pd.DataFrame(encoded_features, columns=encoder.get_feature_names_out(['textuur']))

    # Concatenate the original DataFrame with the encoded DataFrame
    df_SIM_NFGMR = pd.concat([df_SIM_NFGMR, encoded_df], axis=1)


df_predict_CP = predict_RF_lnp(df_arr_res=df_SIM_NFGMR, pred_param="NFGMR", var="CP") #pred_param: LNP column to boost (NFGMR, THARV or TAGBG), and compare eventual RF-prediction to

df_predict_CP = df_predict_CP.rename(columns={'NFGMR': 'LNP_raw', 'prediction': 'Stacked_RF'})
df_predict_CP['LNP_raw'] *= 6.25  # Convert N% to CP
df_predict_CP = df_predict_CP.drop(columns=['sd']).melt(id_vars=['Year', 'Exp_ID', 'observations'], value_vars=['LNP_raw', 'Stacked_RF'])
outdir = os.path.join(current_path, "RF_evaluation_CP",)
os.makedirs(outdir, exist_ok=True)

evaluation_metrics(df_predict_CP, outdir)


# DM yield
CY_path = os.path.join(current_path, "grassat/input/random_forest_training/SIM_DM_cut_yield_THARV.csv")#
df_SIM_CY = pd.read_csv(CY_path)# cut yield, mostly ILVO data
CYest_path = os.path.join(current_path, "grassat/input/random_forest_training/SIM_DM_Yield_Est_TAGBG.csv")#
df_SIM_Yest = pd.read_csv(CYest_path ) # estimated yield, mostly Inagro & Hooibeekhoeve data
df_SIM_Y = pd.concat([df_SIM_CY, df_SIM_Yest], ignore_index=True)# Combine yield simulations

# Fit and transform the categorical feature
if texture==True:
    encoder = OneHotEncoder()
    encoded_features = encoder.fit_transform(df_SIM_NFGMR[['textuur']]).toarray()
    # Convert the encoded features to a DataFrame
    encoded_df = pd.DataFrame(encoded_features, columns=encoder.get_feature_names_out(['textuur']))

    # Concatenate the original DataFrame with the encoded DataFrame
    df_SIM_NFGMR = pd.concat([df_SIM_NFGMR, encoded_df], axis=1)

df_predict_both = predict_RF_lnp(df_arr_res=df_SIM_Y, pred_param="TAGBG", var="Y")# train & predict RF models on test set

df_predict_both = df_predict_both.rename(columns={'TAGBG': 'LNP_raw', 'prediction': 'Stacked_RF'})
df_predict_both = df_predict_both.drop(columns=['sd']).melt(id_vars=['Year', 'Exp_ID', 'observations'], value_vars=['LNP_raw', 'Stacked_RF'])

outdir = os.path.join(current_path, "RF_evaluation_Y",)
os.makedirs(outdir, exist_ok=True)

evaluation_metrics(df_predict_both, outdir)
